# Copyright (c) Meta Platforms, Inc. and affiliates.
# All rights reserved.

# This source code is licensed under the license found in the
# LICENSE file in the root directory of this source tree.

"""
A minimal training script for DiT.
"""
import torch
# the first flag below was False when we tested this script but True makes A100 training a lot faster:
torch.backends.cuda.matmul.allow_tf32 = True
torch.backends.cudnn.allow_tf32 = True
import torch.distributed as dist
from torch.nn.parallel import DistributedDataParallel as DDP
from torch.utils.data import Dataset, DataLoader
from torch.utils.data.distributed import DistributedSampler
from torchvision.datasets import ImageFolder
from torchvision import transforms
import numpy as np
from collections import OrderedDict
from PIL import Image
from copy import deepcopy
from glob import glob
from time import time
import argparse
import logging
import os
from accelerate import Accelerator

from models import DiT_models
from diffusion import create_diffusion
from diffusers.models import AutoencoderKL


#################################################################################
#                             Training Helper Functions                         #
#################################################################################

@torch.no_grad()
def update_ema(ema_model, model, decay=0.9999):
    """
    Step the EMA model towards the current model.
    """
    ema_params = OrderedDict(ema_model.named_parameters())
    model_params = OrderedDict(model.named_parameters())

    for name, param in model_params.items():
        name = name.replace("module.", "")
        # TODO: Consider applying only to params that require_grad to avoid small numerical changes of pos_embed
        ema_params[name].mul_(decay).add_(param.data, alpha=1 - decay)


def requires_grad(model, flag=True):
    """
    Set requires_grad flag for all parameters in a model.
    """
    for p in model.parameters():
        p.requires_grad = flag


def create_logger(logging_dir):
    """
    Create a logger that writes to a log file and stdout.
    """
    logging.basicConfig(
        level=logging.INFO,
        format='[\033[34m%(asctime)s\033[0m] %(message)s',
        datefmt='%Y-%m-%d %H:%M:%S',
        handlers=[logging.StreamHandler(), logging.FileHandler(f"{logging_dir}/log.txt")]
    )
    logger = logging.getLogger(__name__)
    return logger


def center_crop_arr(pil_image, image_size):
    """
    Center cropping implementation from ADM.
    https://github.com/openai/guided-diffusion/blob/8fb3ad9197f16bbc40620447b2742e13458d2831/guided_diffusion/image_datasets.py#L126
    """
    while min(*pil_image.size) >= 2 * image_size:
        pil_image = pil_image.resize(
            tuple(x // 2 for x in pil_image.size), resample=Image.BOX
        )

    scale = image_size / min(*pil_image.size)
    pil_image = pil_image.resize(
        tuple(round(x * scale) for x in pil_image.size), resample=Image.BICUBIC
    )

    arr = np.array(pil_image)
    crop_y = (arr.shape[0] - image_size) // 2
    crop_x = (arr.shape[1] - image_size) // 2
    return Image.fromarray(arr[crop_y: crop_y + image_size, crop_x: crop_x + image_size])


class CustomDataset(Dataset):
    def __init__(self, features_dir, labels_dir):
        self.features_dir = features_dir
        self.labels_dir = labels_dir

        self.features_files = sorted(os.listdir(features_dir))
        self.labels_files = sorted(os.listdir(labels_dir))

    def __len__(self):
        assert len(self.features_files) == len(self.labels_files), \
            "Number of feature files and label files should be same"
        return len(self.features_files)

    def __getitem__(self, idx):
        feature_file = self.features_files[idx]
        label_file = self.labels_files[idx]

        features = np.load(os.path.join(self.features_dir, feature_file))
        labels = np.load(os.path.join(self.labels_dir, label_file))
        return torch.from_numpy(features), torch.from_numpy(labels)
    
def calculate_2_inf_norms(dit_model):
    max_2_inf_norms = {
        'W_q': 0,
        'W_k': 0,
        'W_v': 0,
        'W_o': 0,
        'W1': 0
    }

    # Iterate through each DiTBlock in the model
    for block in dit_model.blocks:
        # Attention weights extraction
        qkv_weight = block.attn.qkv.weight
        output_weight = block.attn.proj.weight
        hidden_size = qkv_weight.shape[0] // 3  # Assuming the hidden_size is a third of the total qkv weight size

        # Calculate 2-infinity norm for each weight type
        norms_q = qkv_weight[:hidden_size].norm(p=2, dim=1).max().item()
        norms_k = qkv_weight[hidden_size:2*hidden_size].norm(p=2, dim=1).max().item()
        norms_v = qkv_weight[2*hidden_size:3*hidden_size].norm(p=2, dim=1).max().item()
        norms_o = output_weight.norm(p=2, dim=1).max().item()

        # Update max norms
        max_2_inf_norms['W_q'] = max(max_2_inf_norms['W_q'], norms_q)
        max_2_inf_norms['W_k'] = max(max_2_inf_norms['W_k'], norms_k)
        max_2_inf_norms['W_v'] = max(max_2_inf_norms['W_v'], norms_v)
        max_2_inf_norms['W_o'] = max(max_2_inf_norms['W_o'], norms_o)

        # MLP weights 2-infinity norm
        norms_W1 = block.mlp.fc1.weight.norm(p=2, dim=1).max().item()
        max_2_inf_norms['W1'] = max(max_2_inf_norms['W1'], norms_W1)

    return max_2_inf_norms

def calculate_2_norms(dit_model):
    max_2_norms = {
        'W_q': 0,
        'W_k': 0,
        'W_v': 0,
        'W_o': 0,
        'W1': 0
    }

    # Iterate through each DiTBlock in the model
    for block in dit_model.blocks:
        # Attention weights extraction
        qkv_weight = block.attn.qkv.weight
        output_weight = block.attn.proj.weight
        hidden_size = qkv_weight.shape[0] // 3  # Assuming the hidden_size is a third of the total qkv weight size

        # Calculate 2-norm (largest singular value) for each weight type
        U_q, S_q, V_q = torch.svd(qkv_weight[:hidden_size])
        U_k, S_k, V_k = torch.svd(qkv_weight[hidden_size:2*hidden_size])
        U_v, S_v, V_v = torch.svd(qkv_weight[2*hidden_size:3*hidden_size])
        U_o, S_o, V_o = torch.svd(output_weight)

        norms_q = S_q.max().item()
        norms_k = S_k.max().item()
        norms_v = S_v.max().item()
        norms_o = S_o.max().item()

        # Update max norms
        max_2_norms['W_q'] = max(max_2_norms['W_q'], norms_q)
        max_2_norms['W_k'] = max(max_2_norms['W_k'], norms_k)
        max_2_norms['W_v'] = max(max_2_norms['W_v'], norms_v)
        max_2_norms['W_o'] = max(max_2_norms['W_o'], norms_o)

        # MLP weights 2-norm
        U_W1, S_W1, V_W1 = torch.svd(block.mlp.fc1.weight)
        norms_W1 = S_W1.max().item()
        max_2_norms['W1'] = max(max_2_norms['W1'], norms_W1)

    return max_2_norms


#################################################################################
#                                  Testing                               #
#################################################################################

def main(args):
    """
    Trains a new DiT model.
    """
    assert torch.cuda.is_available(), "Training currently requires at least one GPU."

    # Setup accelerator:
    accelerator = Accelerator()
    device = accelerator.device

    # Setup an experiment folder:
    if accelerator.is_main_process:
        os.makedirs(args.results_dir, exist_ok=True)  # Make results folder (holds all experiment subfolders)
        experiment_index = len(glob(f"{args.results_dir}/*"))
        model_string_name = args.model.replace("/", "-")  # e.g., DiT-XL/2 --> DiT-XL-2 (for naming folders)
        experiment_dir = f"{args.results_dir}/{experiment_index:03d}-{model_string_name}"  # Create an experiment folder
        checkpoint_dir = f"{experiment_dir}/checkpoints"  # Stores saved model checkpoints
        os.makedirs(checkpoint_dir, exist_ok=True)
        logger = create_logger(experiment_dir)
        logger.info(f"Experiment directory created at {experiment_dir}")

    # latent_size = args.image_size // 8
    latent_size = args.image_size
    model = DiT_models[args.model](
        input_size=latent_size,
        num_classes=args.num_classes,
        latent=args.latent
    )
    # Note that parameter initialization is done within the DiT constructor
    model = model.to(device)
    ema = deepcopy(model).to(device)  # Create an EMA of the model for use after training
    requires_grad(ema, False)
    diffusion = create_diffusion(timestep_respacing="")  # default: 1000 steps, linear noise schedule
    # vae = AutoencoderKL.from_pretrained(f"stabilityai/sd-vae-ft-{args.vae}").to(device)
    if accelerator.is_main_process:
        logger.info(f"DiT Parameters: {sum(p.numel() for p in model.parameters()):,}")

    # Setup data:
    features_dir = f"{args.feature_path}/features"
    labels_dir = f"{args.feature_path}/labels"
    dataset = CustomDataset(features_dir, labels_dir) 
    loader = DataLoader(
        dataset,
        batch_size=int(args.global_batch_size // accelerator.num_processes),
        shuffle=True,
        num_workers=args.num_workers,
        pin_memory=True,
        drop_last=True
    )

    # Load ema model
    checkpoint = torch.load(args.model_path)
    ema.load_state_dict(checkpoint['ema'])
    ema.eval()  

    # Variables for monitoring/logging purposes:
    train_steps = 0
    log_steps = 0
    running_loss = 0
    start_time = time()
        
    for x, y in loader:
        x = x.to(device)
        y = y.to(device)
        x = x.squeeze(dim=1)
        y = y.squeeze(dim=1)
        if args.random_time:
            t = torch.randint(0, diffusion.num_timesteps, (x.shape[0],), device=device)
        else:
            t = torch.full((x.shape[0],), args.constant_timestep, dtype=torch.float, device=device)
        model_kwargs = dict(y=y)
        loss_dict = diffusion.training_losses(ema, x, t, model_kwargs)
        loss = loss_dict["loss"].mean()

        # Log loss values:
        running_loss += loss.item()
        log_steps += 1
        train_steps += 1
        
    avg_loss = torch.tensor(running_loss / log_steps, device=device)
    avg_loss = avg_loss.item() / accelerator.num_processes

    if accelerator.is_main_process:
        logger.info(f"EMA testing loss {avg_loss}")
        
    max_norms = calculate_2_inf_norms(ema)
    
    if accelerator.is_main_process:
        for name, norm in max_norms.items():
            logger.info(f"Max 2-infinity norms. {name}: {norm}")
            
    max_norms = calculate_2_norms(ema)
    
    if accelerator.is_main_process:
        for name, norm in max_norms.items():
            logger.info(f"Max 2 norms. {name}: {norm}")
        
    # Load model
    model.load_state_dict(checkpoint['model'])
    model.eval()  

    # Variables for monitoring/logging purposes:
    train_steps = 0
    log_steps = 0
    running_loss = 0
    start_time = time()
        
    for x, y in loader:
        x = x.to(device)
        y = y.to(device)
        x = x.squeeze(dim=1)
        y = y.squeeze(dim=1)
        t = torch.randint(0, diffusion.num_timesteps, (x.shape[0],), device=device)
        model_kwargs = dict(y=y)
        loss_dict = diffusion.training_losses(ema, x, t, model_kwargs)
        loss = loss_dict["loss"].mean()

        # Log loss values:
        running_loss += loss.item()
        log_steps += 1
        train_steps += 1
        
    avg_loss = torch.tensor(running_loss / log_steps, device=device)
    avg_loss = avg_loss.item() / accelerator.num_processes

    if accelerator.is_main_process:
        logger.info(f"Model testing loss {avg_loss}")
        
    max_norms = calculate_2_inf_norms(model)
    
    if accelerator.is_main_process:
        for name, norm in max_norms.items():
            logger.info(f"Model Max 2-infinity norms. {name}: {norm}")
            
    max_norms = calculate_2_norms(model)
    
    if accelerator.is_main_process:
        for name, norm in max_norms.items():
            logger.info(f"Model Max 2 norms. {name}: {norm}")


if __name__ == "__main__":
    # Default args here will train DiT-XL/2 with the hyperparameters we used in our paper (except training iters).
    parser = argparse.ArgumentParser()
    parser.add_argument("--feature-path", type=str, default="features")
    parser.add_argument("--results-dir", type=str, default="results")
    parser.add_argument("--model", type=str, choices=list(DiT_models.keys()), default="DiT-XL/2")
    parser.add_argument("--model-path", type=str, default="results")
    parser.add_argument("--image-size", type=int, default=256)
    parser.add_argument("--num-classes", type=int, default=10)
    parser.add_argument("--global-batch-size", type=int, default=256)
    parser.add_argument("--global-seed", type=int, default=0)
    parser.add_argument("--num-workers", type=int, default=4)
    parser.add_argument("--latent", type=bool, default=False)
    parser.add_argument("--random_time", type=bool, default=True)
    parser.add_argument("--constant_timestep", type=int, default=5)
    args = parser.parse_args()
    main(args)
